import os
import time
from datetime import date
from math import ceil

from lxml import etree

from r2gg._osm_building import writeNode, writeWay, writeWayNds, writeRes, writeWayTags
from r2gg._osm_to_pbf import osm_to_pbf
from r2gg._sql_building import getQueryByTableAndBoundingBox
from r2gg._database import DatabaseManager


def pivot_to_osm(config, source, db_configs, database: DatabaseManager, logger, output_is_pbf=False):
    """
    Fonction de conversion depuis la bdd pivot vers le fichier osm puis pbf le cas échéant

    Parameters
    ----------
    config: dict
        dictionnaire correspondant à la configuration décrite dans le fichier passé en argument
    source: dict
    db_configs: dict
        dictionnaire correspondant aux configurations des bdd
    database: r2gg.DatabaseManager
        gestionnaire de connexion et d'exécution de la base de la bdd
    logger: logging.Logger
    """
    logger.info("Convert pivot to OSM format for a source")

    # Récupération de la date d'extraction
    work_dir_config = config['workingSpace']['directory']

    # Get extraction date from file or use current date
    date_file = os.path.join(work_dir_config, "r2gg.date")
    if os.path.exists(date_file):
        f = open(date_file, "r")
        extraction_date = f.read()
        f.close()
    else:
        extraction_date = date.today().strftime("%Y-%m-%d")

    source_db_config = db_configs[source['mapping']['source']['baseId']]
    input_schema = source_db_config.get('schema')

    last_value_nodes_query = f"select last_value from {input_schema}.nodes_id_seq"
    vertexSequence, _ = database.execute_select_fetch_one(last_value_nodes_query, show_duration=True)
    vertexSequence = vertexSequence[0]
    logger.info(vertexSequence)

    last_value_edges_query = f"select last_value from {input_schema}.edges_id_seq"
    edgeSequence, _ = database.execute_select_fetch_one(last_value_edges_query, show_duration=True)
    edgeSequence = edgeSequence[0]
    logger.info(edgeSequence)

    logger.info("Starting conversion from pivot to OSM")
    start_time = time.time()

    filename = os.path.join(work_dir_config, source['id'] + ".osm")
    logger.info("OSM file to write : " + filename)

    os.makedirs(os.path.dirname(filename) or '.', exist_ok=True)
    try:
        with etree.xmlfile(filename, encoding='utf-8', close=True) as xf:
            xf.write_declaration()
            attribs = {"version": "0.6", "generator": "r2gg"}
            with xf.element("osm", attribs):

                # Récupération du nombre de nodes
                number_of_nodes_query = f"SELECT COUNT(*) as cnt FROM {input_schema}.nodes"
                row, _ = database.execute_select_fetch_one(number_of_nodes_query, show_duration=True)
                nodesize = row["cnt"]

                # Ecriture des nodes
                batchsize = 500000
                offset = 0
                logger.info(f"Writing nodes: {nodesize} ways to write")
                st_nodes = time.time()
                while offset < nodesize:
                    sql_query_nodes = getQueryByTableAndBoundingBox(f'{input_schema}.nodes', source['bbox'])
                    sql_query_nodes += " LIMIT {} OFFSET {}".format(batchsize, offset)
                    offset += batchsize
                    logger.info("Writing nodes")
                    for row, count in database.execute_select_fetch_multiple(sql_query_nodes, show_duration=True):
                        nodeEl = writeNode(row, extraction_date)
                        xf.write(nodeEl, pretty_print=True)

                    logger.info("%s / %s nodes ajoutés" % (offset, nodesize))
                et_nodes = time.time()
                logger.info("Writing nodes ended. Elapsed time : %s seconds." % (et_nodes - st_nodes))

                # Récupération du nombre de ways
                sql_query_edges_count = f"SELECT COUNT(*) as cnt FROM {input_schema}.edges"
                row, _ = database.execute_select_fetch_one(sql_query_edges_count, show_duration=True)
                edgesize = row["cnt"]

                # Ecriture des ways
                batchsize = 300000
                offset = 0
                logger.info(f"Writing ways: {edgesize} ways to write")
                st_edges = time.time()
                while offset < edgesize:
                    sql_query_edges = getQueryByTableAndBoundingBox(f'{input_schema}.edges', source['bbox'], ['*',
                                                                                                              f'{input_schema}.inter_nodes(geom) as internodes'])
                    sql_query_edges += " LIMIT {} OFFSET {}".format(batchsize, offset)
                    offset += batchsize
                    for row, count in database.execute_select_fetch_multiple(sql_query_edges, show_duration=True):
                        wayEl = writeWay(row, extraction_date)
                        for node in row['internodes']:
                            vertexSequence = vertexSequence + 1
                            node['id'] = vertexSequence
                            nodeEl = writeNode(node, extraction_date)
                            xf.write(nodeEl, pretty_print=True)
                        wayEl = writeWayNds(wayEl, row, row['internodes'])
                        wayEl = writeWayTags(wayEl, row)
                        xf.write(wayEl, pretty_print=True)

                    logger.info("%s / %s ways ajoutés" % (offset, edgesize))
                et_edges = time.time()
                logger.info("Writing ways ended. Elapsed time : %s seconds." % (et_edges - st_edges))

                # Ecriture des restrictions
                sql_query_non_comm = f"select * from {input_schema}.non_comm"
                logger.info("Writing restrictions")
                st_execute = time.time()
                i = 1
                for row, count in database.execute_select_fetch_multiple(sql_query_non_comm, show_duration=True):
                    if row['common_vertex_id'] == -1:
                        i += 1
                        continue
                    ResEl = writeRes(row, i, extraction_date)
                    xf.write(ResEl, pretty_print=True)
                    if (i % ceil(count / 10) == 0):
                        logger.info("%s / %s restrictions ajoutés" % (i, count))
                    i += 1

                et_execute = time.time()
                logger.info("Writing restrictions ended. Elapsed time : %s seconds." % (et_execute - st_execute))

    except etree.SerialisationError:
        logger.warning("WARNING: XML file not closed properly (lxml.etree.SerialisationError)")

    end_time = time.time()
    logger.info("Conversion from pivot to OSM ended. Elapsed time : %s seconds." % (end_time - start_time))

    # osm2pbf : Gestion du format osm.pbf
    if output_is_pbf:
        osm_to_pbf(filename, filename + '.pbf', logger)



import time
from os import getenv

import psycopg2
from psycopg2 import OperationalError, DatabaseError, InterfaceError
from psycopg2.extras import DictCursor
import logging

TIMEOUT = int(getenv("SQL_STATEMENT_TIMEOUT", 0))
RETRY = int(getenv("SQL_STATEMENT_RETRY_ATTEMPTS", 3))
DELAY = int(getenv("SQL_DELAY_BETWEEN_STATEMENTS", 30))


def database_retry_decorator(func):
    def wrapper(self, *args, **kwargs):
        attempt = 1
        while attempt <= RETRY:
            try:
                self.ensure_connection()
                yield from func(self, *args, **kwargs)
                return

            except (OperationalError, DatabaseError, InterfaceError) as e:
                if attempt >= RETRY:
                    self.logger.error(f"Query failed after {RETRY} attempts: {str(e).rstrip()}")
                    raise e

                self.logger.error(
                    f"Attempt {attempt}/{RETRY} failed ({str(e).rstrip()}), retrying in {DELAY} seconds"
                )
                time.sleep(DELAY)
                attempt += 1
                try:
                    if self._connection:
                        self._connection.rollback()
                except Exception as e:
                    self.logger.error(f"Connection rollback failed {str(e).rstrip()}")
        return

    return wrapper


class DatabaseManager:
    def __init__(self, db_configs, logger):
        self.logger = logger
        self._work_db_config = db_configs
        self._connection = self.connect_working_db()

    def connect_working_db(self):
        """
        Fonction de connexion à la BDD de travail

        Parameters
        ----------
        config: dict
            dictionnaire correspondant à la configuration décrite dans le fichier passé en argument
        db_configs: dict
            dictionnaire correspondant aux configurations des bdd
        Returns
        -------
        connection: psycopg2.connection
            connection à la bdd de travail

        """
        # Récupération des paramètres de la bdd
        host = self._work_db_config.get("host")
        dbname = self._work_db_config.get("database")
        user = self._work_db_config.get("user")
        password = self._work_db_config.get("password")
        port = self._work_db_config.get("port")
        connect_args = "host=%s dbname=%s user=%s password=%s port=%s" % (host, dbname, user, password, port)

        self.logger.info("Connecting to work database")
        connection = psycopg2.connect(connect_args)
        connection.set_client_encoding("UTF8")

        return connection

    def disconnect_working_db(self):
        """
        Fonction de connexion à la BDD de travail

        Parameters
        ----------
        connection: psycopg2.connection
            connection à la bdd de travail
        logger: logging.Logger
        """
        if self._connection:
            self._connection.close()
            self.logger.info("Connection to work database closed")

    def ensure_connection(self):
        """
        Ensure the connection is alive; reconnect if needed.
        """
        try:
            if self._connection is None or getattr(self._connection, "closed", 1) != 0:
                self.logger.info("Connection is closed or missing; reconnecting")
                self._connection = self.connect_working_db()
            else:
                with self._connection.cursor() as cur:
                    cur.execute("SELECT 1")
        except Exception as e:
            self.logger.error(
                f"Something is wrong with the connection: {str(e).rstrip()}; reconnecting in {DELAY} seconds")
            self.disconnect_working_db()
            time.sleep(DELAY)
            self._connection = self.connect_working_db()

    def execute_select_query(self, cursor, query, show_duration):
        if TIMEOUT:
            cursor.execute("SET statement_timeout = %s", (1000 * TIMEOUT,))  # timeout in milliseconds

        if show_duration:
            self.logger.info("SQL: {}".format(query))
            st_execute = time.time()
            cursor.execute(query)
            et_execute = time.time()
            self.logger.info("Execution ended. Elapsed time : %s seconds." % (et_execute - st_execute))
        else:
            cursor.execute(query)

    @database_retry_decorator
    def execute_select_fetch_multiple(self, query, batchsize=1, show_duration=False):
        with self._connection.cursor(cursor_factory=DictCursor) as cursor:
            self.execute_select_query(cursor, query, show_duration)
            rows = cursor.fetchmany(batchsize)
            count = cursor.rowcount
            while rows:
                if batchsize == 1:
                    rows = rows.pop()
                yield rows, count
                rows = cursor.fetchmany(batchsize)
            self._connection.commit()
            return

    # the method below should be used as a generator function otherwise use execute_update
    @database_retry_decorator
    def execute_update_query(self, query, params=None, isolation_level=None, show_duration=False):
        if show_duration :
            self.logger.info("SQL: {}".format(query))
        st_execute = time.time()
        with self._connection.cursor(cursor_factory=DictCursor) as cursor:
            old_isolation_level = self._connection.isolation_level
            if isolation_level is not None:
                self._connection.set_isolation_level(isolation_level)
            cursor.execute(query, params)
            self._connection.commit()
            if show_duration:
                et_execute = time.time()
                self.logger.info("Execution ended. Elapsed time : %s seconds." % (et_execute - st_execute))
            self._connection.set_isolation_level(old_isolation_level)
        yield  # the decorator database_retry_decorator only supports generators
        return

    def execute_update(self, query, params=None, isolation_level=None):
        next(self.execute_update_query(query, params=params, isolation_level=isolation_level), None)

    def execute_select_fetch_one(self, query, show_duration=False):
        gen = self.execute_select_fetch_multiple(query, 1, show_duration)
        row, count = next(gen, (None, None))
        return row, count



#! python3  # noqa: E265

"""Main CLI entrypoint."""

# Package
from r2gg._configure import configure
from r2gg._main import sql_convert, pgr_convert, osm_convert, osrm_convert, valhalla_convert, write_road2_config
from r2gg._database import DatabaseManager

# ############################################################################
# ########## MAIN ################
# ################################
def sql2pivot():
    config, resource, db_configs, logger = configure()
    database = DatabaseManager(db_configs[config["workingSpace"]["baseId"]], logger)
    sql_convert(config, resource, db_configs, database, logger)
    database.disconnect_working_db()

def pivot2pgrouting():
    config, resource, db_configs, logger = configure()
    database = DatabaseManager(db_configs[config["workingSpace"]["baseId"]], logger)
    pgr_convert(resource, db_configs, database, logger)
    database.disconnect_working_db()

def pivot2osm():
    config, resource, db_configs, logger = configure()
    database = DatabaseManager(db_configs[config["workingSpace"]["baseId"]], logger)
    osm_convert(config, resource, db_configs, database, logger)
    database.disconnect_working_db()

def osm2osrm():
    config, resource, _, logger = configure()
    osrm_convert(config, resource, logger)

def osm2valhalla():
    config, resource, _, logger = configure()
    valhalla_convert(config, resource, logger)

def road2config():
    config, resource, _, logger = configure()
    write_road2_config(config, resource, logger)

def main():
    """Main CLI entrypoint.
    """
    config, resource, db_configs, logger = configure()
    sql2pivot()
    if resource['type'] in ['pgr', 'smartpgr']:
        pivot2pgrouting()
    elif resource['type'] == 'osrm':
        pivot2osm()
        osrm_convert(config, resource, logger)
    elif resource['type'] == 'valhalla':
        pivot2osm()
        valhalla_convert(config, resource, logger)
    else:
        raise ValueError("Wrong resource type, should be in ['pgr',osrm','valhalla','smartpgr']")
    write_road2_config(config, resource, logger)

# -- Stand alone execution
if __name__ == "__main__":
    main()  # required by unittest



import argparse
import logging
import os

import psycopg2

from r2gg._read_config import config_from_path

# Définition des niveaux de log
LEVELS = {
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG
}


def configure():
    """
    Fonction de lecture du fichier de configuration passé en argument

    Returns
    -------
    config: dict
        dictionnaire correspondant à la configuration décrite dans le fichier passé en argument
    resource: dict
        dictionnaire correspondant à la resource décrite dans le fichier passé en argument
    db_configs: dict
        dictionnaire correspondant aux configurations des bdd
    logger: logging.Logger
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('config_file_path', type=str)
    args = parser.parse_args()
    config_path = args.config_file_path

    # Récupération de l'objet 'génération' qui contient toute la config
    config = config_from_path(config_path)['generation']

    # Récupération de la configuration du log
    logs_config = config_from_path(config['general']['logs']['configFile'])

    # Gestion du fichiers de logs non spécifié
    try:
        logs_file = logs_config['filename']
    except KeyError:
        logs_file = '/dev/null'

    # Configuration du module logging
    logging.basicConfig(
        format='%(asctime)s %(message)s',
        level=LEVELS[logs_config['level'].upper()],
        handlers=[
            logging.FileHandler(logs_file),
            logging.StreamHandler()
        ])

    # Initialisation du logger
    logger = logging.getLogger(__name__)
    logger.info("Log initialized")

    # Todo : Créer une fonction qui vérifie la configuration
    db_configs = {}
    # Configuration des bases de données précisées dans la config
    for base in config['bases']:
        if base['type'] == 'bdd':
            db_configs[base['id']] = config_from_path(base['configFile'])
            db_configs[base['id']].update({"schema": base['schema']})

    #  Récupération de l'objet permettant de générer la ressource
    resource = config['resource']

    # Création de l'espace de travail
    if not os.path.exists(config['workingSpace']['directory']):
        os.makedirs(config['workingSpace']['directory'])

    return config, resource, db_configs, logger
